import tkinter as tk
from tkinter import filedialog
import os, shutil,sys
sys.path.append(os.path.dirname(__file__))
from load_save_helpers import create_default_params, save_parameters, load_parameters
from BBHP_function import param_types, BBHP

base_param_name = 'base_parameters'

def runBB():
    
    
    file_path = get_filepath()
    param = load_parameters(file_path+base_param_name,param_types)

    BBHP(file_path,param)#, train_img_count = train_img_count)#,dt_string='22_06_06_20_44_28');#,train_img_count=train_img_count,test_img_count=test_img_count,dt_string=dt_string)


# returns images of a list of filenames for a given folder
def createdir(example=0):
    
    # request filepath from user
    file_path = get_filepath()

    # if pre-loaded example, images, masks, etc will be added
    if example == 1:
        folder_name = 'PhotoelasticDisks'
        
    elif not example == 0:
        print('[BB] Bad input, no such example #'+example+'.')
        return
    
    else: # example == 0
        folder_name = input('[INPUT] Enter folder name for new project: ')
        
        
    full_filename = file_path+folder_name
    if os.path.isdir(full_filename):
        print('[BB] Desired new folder already exists! Nohting has been changed.')
        return
   
    
    
    
    # given example must just copy existing directory in package
    if example >0:
        tocopy_dir = os.path.dirname(__file__)+'/'+folder_name

        shutil.copytree(tocopy_dir, full_filename)
       
    # non-example, just create folders and default parameter file
    else:
        os.mkdir(full_filename)
        os.mkdir(full_filename+'/train_images')
        os.mkdir(full_filename+'/test_images')
        os.mkdir(full_filename+'/predict_images')
        os.mkdir(full_filename+'/masks')
        os.mkdir(full_filename+'/areas_of_interest')
        
    param = create_default_params();

    save_parameters(full_filename+'/'+base_param_name,list(param.keys()),list(param.values()))

        
    print('Created and populated '+full_filename)

    
def get_filepath():
    # request filepath from user
    root = tk.Tk()
    root.withdraw()
    file_path = filedialog.askdirectory()
    return file_path + '/'
    


import numpy as np
import tensorflow as tf
from tensorflow.keras.utils import to_categorical
from scipy.signal import convolve2d
import matplotlib.pyplot as plt
import copy
import cv2
import gc
from load_save_helpers import find_matching_masks, list_imgs_and_sizes, get_image_list, load_image_list, distance_maker, primelist, circ_region


def_neighborhood_radius = 1.01
def_particle_border_mult = 1, 
def_two_particle_mult = 1, 
def_img_edge_mult = 1,

def_dist_max = 10;




#################################################################################################################       
#################################################################################################################      
########################    CreateGenerator from param/filepath  ################################################ 
#################################################################################################################   
#################################################################################################################   

def CreateGenerator(filepath, subfolder, param, want=None, count=-1, train_test_predict=0, index = -1, chatter = False):
    
    # determine type of generator
    if train_test_predict == 0: # training generator: include manipulations of data and shuffle data
        flip,rotate,randomize = param['flips'],param['rotations'],True
    elif train_test_predict == 1 or train_test_predict == 2: # testing/prediction generator: leave data alone 
        flip,rotate,randomize = False,False,False
    else:
        raise Exception('train_test_predict parameter must be 0 (training), 1 (testing), or 2 (predicting).')

    # create generator object
    dataGenerator = HPinputGenerator(S_half = param['S_half'], \
                                      batch_size = param['batch_size'],\
                                      scales=param['scales'], \
                                      scalefactor=param['scalefactor'], \
                                      flip=flip, rotate=rotate, randomize=randomize,\
                                      dist_max = param['dist_max'])
    
    # load images specified in inputs
    img_names = get_image_list(filepath+subfolder, want=want, count=count)
    
    if train_test_predict == 0 and len(img_names)==0:
        raise Exception('Attempt to create training generator without any images!')
        
    if index!=-1:
        if index>=len(img_names):
            raise IndexError('Index outside of want list')
        else:
            img_names = [img_names[index]]
    imgs = load_image_list(filepath+subfolder, img_names, RGB_to_gray=param['images_to_grayscale'], chatter = chatter)
    if chatter:
        print('[BB] -- '+subfolder+' images: '+list_imgs_and_sizes(img_names,imgs))
        
    # if training or testing, we need masks, find names
    if train_test_predict < 2: 
        mask_folder = filepath+'masks'
        mask_names = find_matching_masks(img_names, get_image_list(mask_folder))
        masks = load_image_list(mask_folder,mask_names, integerize=True,RGB_to_gray = True, segment = True)
        if chatter:
            print('[BB] -- '+subfolder+' masks: '+list_imgs_and_sizes(mask_names,masks))
    else:
        masks = [None for k in imgs]
        
    # look for AOIs, load them (allow only 1's and 0's)
    AOI_folder = filepath+'areas_of_interest'
    AOI_names = find_matching_masks(img_names, get_image_list(AOI_folder), raise_issues=False)
    AOIs = load_image_list(AOI_folder, AOI_names, binarize=True)  
    
    # for all non-defined AOIs, assume entire image is fair game
    for k,img in enumerate(imgs): 
        if AOIs[k] is None:
            AOIs[k] = np.ones(np.shape(img))

    if chatter:
        print('[BB] -- '+subfolder+' AOIs: '+list_imgs_and_sizes(AOI_names,AOIs))
    
    # Enforce specified fraction of data if training set
    if train_test_predict == 0:
        AOIs = fractionalize_AOI(masks, AOIs, param['fraction'])

        
    # if there are masks and track_outies flag is true, alert user and invert masks
    if param['track_outies'] and np.sum([not m is None for m in masks])>0:
        if chatter:
            print('Inverting masks to track zero-valued pixels.')
        masks = [1*(m == 0) if (not m is None) else None for m in masks]
    
    # create HP inputs
    for k,img in enumerate(imgs):
        dataGenerator.add_ID_img(img, mask=masks[k], AOI=AOIs[k], neighborhood_radius=param['HP_neighborhood_radius'], particle_border_mult=param['HP_particle_border_mult'], two_particle_mult=param['HP_two_particle_mult'],  img_edge_mult = param['HP_img_edge_mult'])                   

    return dataGenerator, img_names, copy.deepcopy(masks)


#################################################################################################################       
#################################################################################################################      
########################    Generator Class   ################################################################### 
#################################################################################################################   
#################################################################################################################   




    
# generator class that generates img - label pairs from a given img, row/col arrays, snapshot half size, and labels
class BBinputGenerator(tf.keras.utils.Sequence):
  
    def __init__(self, S_half=20, batch_size=32, flip=True, rotate=True, scales=1, scalefactor=2, randomize = True, neighborhood_radius=def_neighborhood_radius,dist_max = def_dist_max):
        
       
        self.S2 = S_half
        self.S = S_half*2+1
        self.batch_size = batch_size
        self.flip = flip
        self.rotate = rotate
        self.mult = 1*(1+3*rotate)*(1+flip) # flipping doubles, rotating 4x's unique images
        self.randomize = randomize
        self.neighborhood_radius = neighborhood_radius # neighborhood for weighting and erosion if needed
        self.getcount = 0
        self.maxgetcount = 100 # calls garbage collection every maxgetcount calls to __getitem__
        self.outstack = None
        self.dist_max = dist_max
        self.scales = scales
        self.scalefactor = scalefactor
        self.padding = scalefactor**(scales-1)*S_half+1 # this is how far away the image edge must be from AOI
        self.input_shape = None # defined when first image is added
          
        # flag for changing mask to categorical, randomizing order, and adding weights.
        self.data_prepared = False 
        # this is performed on the first call of __getitem__, and undone whenever new data is added (add_img or add_img_vec)
               
        if not int(batch_size/self.mult) == batch_size/self.mult:
            raise Exception('When manipulation (flipping/rotating) is on, batch size must be multiple of these manipulations.')

        self.dim3 = None # no images added yet!

       
        # Placeholder fields to be filled with add_image() method
        self.imgs = [] # first index is image number
        self.img_paddings = []
        
        
        # vectors to create output data (must always keep same order as each other)
        self.img_num = []
        self.rows = []
        self.cols = []
        self.sample_weights = None # to be added by the first __getitem__ call
        self.relative_weights = [] # list of relative weights for innies and outies (e.g. edges are double insides)
        self.label_list = [] # innie vs outie
        self.labels = None # this is in "categorical" terms, to be added by the first __getitem__ call
        
        self.dist_list = [];
        
    
    # Add an image to the dataset (along with mask (innie/outie), AOI) 
    # Uses add_img_vec after processing 2D data.
    def add_img(self,img, mask, AOI=None, weights=None):
         
        img = np.copy(img)
                
        shp = np.shape(img)[:2]
        # reshape to 3D (with 3rd dimension 1), and record 2D shape
        if len(np.shape(img)) == 2:
            img = np.reshape(img,(shp[0],shp[1],1))
        
        # generate dummy mask and total AOI if needed
        if mask is None: # if not provided, we just need a space filler
            mask = np.ones(shp)  
        if AOI is None: # if not provided, everything is in the area of interest
            AOI = np.ones(np.shape(mask))   

        if (not shp == np.shape(mask)) or (not shp == np.shape(AOI)):
            raise Exception('img, mask, AOI, must have the same 2D shape')
                                                                     
        rowmat,colmat = np.mgrid[0:np.size(img,0),0:np.size(img,1)]
        
        # flatten AOI and check for multiple counts on a single pixel.
        # pixels will be repeated in the generator n = AOI[r,c] times
        AOI = np.ndarray.flatten(AOI)
        idx_used = []
        while np.max(AOI)>0:
            AOIbool = AOI>0
            AOI[AOIbool] -=1
            idx_used += list(np.argwhere(AOIbool))

        idx_used = np.ndarray.flatten(np.array(idx_used))
        
        # flatten all matrices, and use AOI to trim/duplicate indicees correctly
        rows = np.ndarray.flatten(rowmat)[idx_used]
        cols = np.ndarray.flatten(colmat)[idx_used]
        labels = np.ndarray.flatten(mask)[idx_used]
        
        if not weights is None:
            weights = np.ndarray.flatten(weights)[idx_used]
        
        # use method below to add the data to the generator!
        self.add_img_vec(img, rows, cols, labels, weights)     
        
        gc.collect()
    
    # Add an image to the dataset (along with rows, columns, and labels of points to generate data from)
    def add_img_vec(self,img, rows, cols, labels, weights=None):
        
        if (not len(rows)==len(labels)) or (not len(cols) == len(labels)):
            raise Exception('rows, cols, and labels must have the same length L and be Lx1')
        
        shp = np.shape(img)
        if len(shp) == 2:
            self.imgs.append(np.copy(img).reshape(shp[0],shp[1],1))
            shp = np.array([shp[0],shp[1],1])
        else:
            self.imgs.append(np.copy(img))
          
      
        if self.dim3 is None:
            self.dim3 = shp[2]
            self.input_shape = self.S,self.S,self.scales*self.dim3

        elif not self.dim3 == shp[2]:
            raise Exception('Added images must have same 3rd dimension: '+str(self.dim3))
            
        if weights is None:
            weights = np.ones(np.shape(rows)).astype(float)
          
        
        # determine how far away the AOI is from the edges of the image 
        # (for rows and columns separately) then add padding to the image
        toppad = max(self.padding-np.min(rows),0)
        bottompad = max(self.padding-(shp[0]-np.max(rows)),0)
        leftpad = max(self.padding-np.min(cols),0)
        rightpad = max(self.padding-(shp[1]-np.max(cols)),0)
        padinstruction = ((toppad, bottompad), (leftpad, rightpad),(0,0)) # 3D 
        
        self.imgs[-1] = np.pad(self.imgs[-1],padinstruction)
        self.img_paddings.append(padinstruction)
        
        rows = np.array(rows)+toppad # correct rows and columns  (labels are unchanged)
        cols = np.array(cols)+leftpad
        
            
        # add blurred version(s) of image to accomodate easy fetching of scaled up images    
        # note that padding happens before blurring. However because blurred images are always subsampled
        # at the same frequency as the blur, this order does not effect the eventual outputs.
        baseimg = np.copy(self.imgs[-1])
        for scale in range(1,self.scales): # for every scale except 1 (powers of scalefactor)
            
            n = self.scalefactor**scale
            newimg = np.zeros(np.shape(baseimg))
            
            # blur at many scales and stack (concatenate) the images along the color axis
            for dim in range(self.dim3):   
                newimg[:,:,dim] = convolve2d(baseimg[:,:,dim],np.ones((n,n))/(n**2), boundary='symm', mode='same')
            
            self.imgs[-1] = np.concatenate([self.imgs[-1],newimg],axis=2)
            
        
        self.rows = np.concatenate([self.rows,rows], axis=0).astype(int)
        self.cols = np.concatenate([self.cols,cols], axis=0).astype(int)
        self.img_num = np.concatenate([self.img_num,[len(self.imgs)-1 for _ in rows]], axis=0).astype(int)
        self.label_list = np.concatenate([self.label_list, labels], axis=0).astype(int)
        self.relative_weights = np.concatenate([self.relative_weights, weights], axis=0).astype(float)
        
        self.data_prepared = False # adding more images makes the entire data set no longer prepped!
    
    
     # Add a mask with automatically generated weights near borders, etc. Can add ID'd mask for higher weighting of regions at  particle contacts.
    def add_weighted_img(self,img, mask, AOI=None, neighborhood_radius=def_neighborhood_radius, particle_border_mult=def_particle_border_mult, two_particle_mult=def_two_particle_mult, img_edge_mult = def_img_edge_mult,weights=None):
        
        # create eroded mask and appropriate weights
        if weights is None:
            if particle_border_mult==1 and two_particle_mult == 1 and img_edge_mult == 1: 
                # no need to do anything fancy if just 1's
                weights = np.copy(AOI)
            else:
                weights, _ = weight_and_erosion_maker(mask, neighborhood_radius=neighborhood_radius, \
                                        particle_border_mult=particle_border_mult, two_particle_mult=two_particle_mult, \
                                        img_edge_mult = img_edge_mult, AOI=AOI)
            # if ID'd mask, make binary
        self.add_img(img, mask>0, AOI=AOI, weights=weights)
    
    # returns a 2D map of relative weights for image idx (with any added padding removed)
    def get_weights(self,idx=0): 
        out = np.zeros(np.shape(self.imgs[idx])[:2]) # same 2D shape as image, no channels.
        want = self.img_num == idx
        out[self.rows[want],self.cols[want]] = self.relative_weights[want]
        p = self.get_padding_array(idx)
        return out[p[0]:p[1],p[2]:p[3]]
    
    # returns a 2D map of innie vs outie mask (with any added padding removed)
    def get_masks(self,idx=0): 
        out = np.zeros(np.shape(self.imgs[idx])[:2]) # same 2D shape as image, no channels.
        want = self.img_num == idx
        out[self.rows[want],self.cols[want]] = self.label_list[want]
        p = self.get_padding_array(idx)
        return out[p[0]:p[1],p[2]:p[3]]
    
    def get_dists(self,idx=0): 
        out = np.zeros(np.shape(self.imgs[idx])[:2]) # same 2D shape as image, no channels.
        want = self.img_num == idx
        out[self.rows[want],self.cols[want]] = self.dist_list[want]
        p = self.get_padding_array(idx)
        return out[p[0]:p[1],p[2]:p[3]]
    
    # returns a 2D map of Area of Interest (AOI) for image idx (with any added padding removed)
    def get_AOI(self,idx=0): 
        out = np.zeros(np.shape(self.imgs[idx])[:2]) # same 2D shape as image, no channels.
        want = self.img_num == idx
        out[self.rows[want],self.cols[want]] = 1
        p = self.get_padding_array(idx)
        return out[p[0]:p[1],p[2]:p[3]]
    
    # returns an input image with any added padding removed
    def get_img(self,idx=0):      
        p = self.get_padding_array(idx)
        return self.imgs[idx][p[0]:p[1],p[2]:p[3],:self.dim3] # cut out scalings
    
    # returns array with indicees required to remove padding from an image
    def get_padding_array(self,idx):
        p = np.ndarray.flatten(np.array(self.img_paddings[idx]))
        p[1] = np.shape(self.imgs[idx])[0]-p[1]
        p[3] = np.shape(self.imgs[idx])[1]-p[3]
        return p
    
    # returns all images with any added padding removed
    def get_all_imgs(self):
        outimgs = []
        for idx in range(np.shape(self.imgs)[0]):
            outimgs.append(self.get_img(idx))
        
        return outimgs
    
    # returns max and minimum value of all images
    def get_img_range(self):
        maxval,minval = np.max(self.get_img(0)),np.min(self.get_img(0))
        for idx in range(1,np.shape(self.imgs)[0]):
            maxval = np.max([maxval,np.max(self.get_img(idx))])
            minval = np.min([minval,np.min(self.get_img(idx))])
        
        return maxval,minval
    
    # normalize all images in this generator -- sends maxval to 0.5 and minval to -0.5
    def normalize_images(self,maxval,minval): 
        for idx in range(np.shape(self.imgs)[0]):
            self.imgs[idx] = (self.imgs[idx]-minval)/(maxval-minval) - 0.5
        
        
        
    # returns an array of 2D predictions for each image in self.imgs, based on the model given.
    def get_prediction(self,model):
        
        #Models with these features will generate multiple predictions per pixel and will not match with the 
        # internal arrays like self.rows/self.cols. It will be massively inefficient and is thus prevented.
        if self.rotate or self.flip:
            raise Exception('This generator has flips or rotates and should not be used as a prediction generator.')
            
        #print('Predicting can be a time-consuming step.')
        
        # do the prediction!     
        outputs = model.predict(self)
        class_vec = outputs[0][:,1] # keep only one side of the class output (innie/outie)
        dist_vec = np.squeeze(outputs[1]) # distance to outie
        

        predictions = []
        for idx in range(np.shape(self.imgs)[0]): # for every image, find predictions
            #print('exporting image #'+str(idx))
            want = self.img_num == idx # identify which elements are for this image

            dummy_img = np.zeros(np.shape(self.imgs[idx])[:2]) 
            dummy_img[self.rows[want],self.cols[want]] = class_vec[want] # place predictions in 2D
            


            # remove padding
            p = np.ndarray.flatten(np.array(self.img_paddings[idx]))
            p[1] = np.shape(self.imgs[idx])[0]-p[1]
            p[3] = np.shape(self.imgs[idx])[1]-p[3]
            
            # add distances to output
            dummy_img2 = np.zeros(np.shape(self.imgs[idx])[:2]) 
            dummy_img2[self.rows[want],self.cols[want]] = dist_vec[want] # place predictions in 2D

            dummy_img = np.stack([dummy_img,dummy_img2],axis=2)
            
            # add to overall output array
            predictions.append(dummy_img[p[0]:p[1],p[2]:p[3],:])
           
        return predictions
        
        
    # randomly mix the order of data
    def shuffle(self):
        
        ordr = np.arange(len(self.rows))
        np.random.shuffle(ordr)

        self.label_list = self.label_list[ordr]
        self.dist_list = self.dist_list[ordr]
        self.rows = self.rows[ordr]
        self.cols = self.cols[ordr]
        self.sample_weights = self.sample_weights[ordr]
        self.relative_weights = self.relative_weights[ordr]
        self.img_num = self.img_num[ordr]
        
 
    # return the number of batches to pull for an epoch of training
    def __len__(self):
        return np.ceil(self.mult*len(self.label_list)/self.batch_size).astype(int)
  
    # return a batch of data. This may include rotations and flips, as well as multiple channels and/or scales
    def __getitem__(self, idx00):
        
        self.getcount += 1
        if self.getcount >= self.maxgetcount:
            gc.collect()
            self.getcount=0
        
        # Check to see if the data is prepped. If not... prep!
        if not self.data_prepared: # happens only on the first call.

            self.enforce_dist_max() # crops label list at self.dist_max
            self.dist_list = copy.deepcopy(self.label_list); # store distances
            
            self.label_list = self.label_list>0        # create binary categorization
            
            # generate weights. Equal total weight for both innies and outies (and this is built to allow more classes)
            self.sample_weights = np.zeros(np.shape(self.rows))
            U = np.unique(self.label_list)
            for k in U:
                want = self.label_list==k # identify elements of this class
                # sum of all elements of each class is equal
                self.sample_weights += np.multiply(self.relative_weights,want)/np.sum(self.relative_weights[want])

            self.sample_weights = self.sample_weights/np.mean(self.sample_weights) # make mean weight = 1

            if self.randomize: # if randomization is in the cards, deal.
                self.shuffle()
        
            self.labels = to_categorical(self.label_list) # make Keras-friendly labels
            
            self.data_prepared = True # data is now prepped.
        
        
        # identify what the unique index and number of unique elements in the batch is (repeats are flips/rots)
        idx0 = idx00/self.mult
        batch = self.batch_size/self.mult
        
        # Indices of the data to be pulled.
        # Repeat self.mult times and sort to give an ordered list [32,32,...,32,83,83...,83,120,...]
        vec = sorted(np.repeat(np.arange((idx0*batch),np.min([(idx0+1)*batch,len(self.labels)]),dtype='int'),self.mult))

        # Scales to pull. If self.scales=1, this will just be 1.
        scalevec = np.power(self.scalefactor,np.arange(self.scales)) # e.g. 1,2,4,8, etc
     
        # create stacked image output. 
        # Images may have multiple channels (e.g. colors) and scales: these are stacked in the same dimension (3 in this stack). 
        # There will also be self.mult repeats of every such color and scale stack (for later rotation/flipping).
        # Output stack dimensions are [self.batch_size,self.S,self.S,self.dim3*self.scales] with [batch] unique elements.

        self.outstack = np.stack([ \
            np.concatenate([ \
                self.imgs[self.img_num[idx]][self.rows[idx]-self.S2*s:self.rows[idx]+self.S2*s+1:s, \
                             self.cols[idx]-self.S2*s:self.cols[idx]+self.S2*s+1:s, \
                             self.dim3*snum:self.dim3*(snum+1)].reshape(self.S,self.S,self.dim3) \
                for s,snum in zip(scalevec,range(self.scales))], axis=2) for idx in vec] ,axis=0)
                     
        # If rotations/flips are used, go through stack and modify unaltered images to create these new data.
        if self.flip or self.rotate:
            multiscaledim3= self.scales*self.dim3

            # rotate and flip depending on idx (which indicates which repeat # of a given environment this is)
            for idx in range(len(vec)):
                # reshape image so that it can use numpy rotation/flip functions
                snap = self.outstack[idx,:,:,:].reshape(self.S,self.S,multiscaledim3)
                
                # rotate 90 degrees for every additional index (stack is sorted by index, so this achieves desired effect)
                if self.rotate:
                    snap = np.rot90(snap,idx%4)
                
                # flip half of the images if self.flip
                if self.flip and (idx%self.mult)>=self.mult/2:
                    snap = np.fliplr(snap)
                    
                # reshape to put back into stack
                self.outstack[idx,:,:,:] = snap.reshape([1,self.S,self.S,multiscaledim3])       
        
        return  np.copy(self.outstack), [np.array(self.labels[vec]),np.array(self.dist_list[vec])], np.array(self.sample_weights[vec])
       
    # Plot sample items created by this generator
    def plot_items(self,count=1,nums=None):
        
        if nums is None:
            nums = (self.__len__()*np.random.rand(count)/self.mult).astype(int)
        else:
            count = len(nums)
            
        c = self.scales+self.rotate+self.flip
        fig, axs = plt.subplots(count,c)

        for r0 in range(count):
            snap,inorout,_ = self.__getitem__(nums[r0]/self.mult)
            

            for s in range(self.scales):
                snap0 = snap[0,:,:,s*self.dim3].reshape(self.S,self.S)
                theplot = axs[r0][s]
                theplot.imshow(snap0,cmap=plt.get_cmap('gray'))
                
                if r0==0:
                    x = str(self.scalefactor**s)
                    theplot.set_title(x+'x')
                    
                if s==0:
                    val = inorout[0][s][1]
                    
                    if val == 1:
                        theplot.set_ylabel('Innie')
                    else:
                        theplot.set_ylabel('Outie')

                            
            if self.rotate: # add one rotation if that's happening
                snap0 = snap[1,:,:,0].reshape(self.S,self.S)
                theplot = axs[r0][c-self.flip-1]
                theplot.imshow(snap0,cmap=plt.get_cmap('gray'))
                if r0==0:
                    theplot.set_title('Rotate')
                        
            if self.flip: # add one flip if that's happening
                snap0 = snap[(1+3*self.rotate),:,:,0].reshape(self.S,self.S)
                theplot = axs[r0][c-1]
                theplot.imshow(snap0,cmap=plt.get_cmap('gray'))
                if r0==0:
                    theplot.set_title('Flip')
                
            for c0 in range(c):
                axs[r0][c0].set_xticks([])
                axs[r0][c0].set_yticks([])
                axs[r0][c0].plot(self.S2,self.S2,'r+')
            
            axs[r0][-1].yaxis.set_label_position("right")
            axs[r0][-1].set_ylabel('#'+str(nums[r0]*self.batch_size))

    
    # Plot sample items created by this generator
    def enforce_dist_max(self):
        self.label_list[self.label_list>self.dist_max] = self.dist_max

        


# creates a BBinputGenerator but with no flipping, rotation, or randomization
class predictionGenerator(BBinputGenerator):
    
    # Removed options that are no-go's in prediction
    def __init__(self, S_half=20, batch_size=1024, scales=1, scalefactor=2):
        super().__init__(S_half=S_half, batch_size=batch_size, flip=False, rotate=False, scales=scales, scalefactor=scalefactor, randomize = False)
        
    # allow for adding an image without a mask (just uses zeros as mask) 
    def add_maskless_img(self,img, AOI=None):      
        mask = np.zeros(np.shape(img)[:2])
        super().add_img(img, mask, AOI=AOI, weights=None)

        
        
# Identical to BBinputGenerator except with add_ID_img() method that erodes ID'd mask and weights appropriately
class HPinputGenerator(BBinputGenerator):
    
    # Removed options that are no-go's in prediction
    def __init__(self, S_half=20, batch_size=32, flip=True, rotate=True, scales=1, scalefactor=2, randomize = True,dist_max=def_dist_max):
       
        super().__init__(S_half=S_half, batch_size=batch_size, flip=flip, rotate=rotate, scales=scales, scalefactor=scalefactor, randomize = randomize,dist_max = dist_max)
        
    # add ID'd mask, which includes generating distances to outies
    def add_ID_img(self, img, mask=None, AOI=None, \
                   neighborhood_radius=def_neighborhood_radius, \
                   particle_border_mult=def_particle_border_mult, \
                   two_particle_mult=def_two_particle_mult, \
                   img_edge_mult = def_img_edge_mult):
        
        if mask is None:
            self.add_img(img, None, AOI=AOI, weights=None)
        
        else:
            # create eroded mask and appropriate weights
            weights, _ = weight_and_erosion_maker(mask, neighborhood_radius=neighborhood_radius, \
                        particle_border_mult=particle_border_mult, two_particle_mult=two_particle_mult, \
                        img_edge_mult = img_edge_mult, AOI=AOI)

            self.add_img(img, distance_maker(mask), AOI=AOI, weights=weights)



#################################################################################################################       
#################################################################################################################      
########################    Helper Functions   ##################################################################
#################################################################################################################   
#################################################################################################################   




# creates a 2D weight map (not balanced for innies vs outies yet) and eroded mask based on an ID'd mask. Mult variables determine additional weighting for regions near the borders of particles, near two particles, and near the image borders (neighborhood defined by r).
def weight_and_erosion_maker(mask, neighborhood_radius=def_neighborhood_radius, particle_border_mult=def_particle_border_mult, two_particle_mult=def_two_particle_mult, img_edge_mult = def_img_edge_mult, AOI=None):

    img2 = np.sqrt(np.array(primelist[np.array(mask,dtype=int)],dtype=float)) # make segmented image using only primes, then take sqrt
    kernel = np.array(circ_region(neighborhood_radius),dtype=np.uint8)

    # convolve image over kernel -- those that catch another sqrt(prime) will become non-integers when multiplied by sqrt(prime) img
    eroded_mask = np.multiply(cv2.filter2D(img2,-1,kernel),img2) 

    # detect non integers - these are innies next to other particles
    twoParticles = np.abs(eroded_mask-np.round(eroded_mask))>.000001  #this number is somewhat arbitrary. Too small is a problem though.

    eroded_mask = cv2.erode(np.array(img2>0,dtype=float),kernel) - twoParticles # erode and subtract particle-particle borders
    eroded_mask[eroded_mask<0] = 0 # any subtractions below 0 from twoParticles are ignored.

    dilated_mask = cv2.dilate(np.array(img2>0,dtype=float),kernel)

    if AOI is None:
        AOI = np.ones(mask.shape[:2])


    # implement all weightings
    weights = np.array(AOI >0,dtype=float) #weights nonzero only in AOI

    weights[(dilated_mask-eroded_mask)>0] *= particle_border_mult
    weights[twoParticles] *= two_particle_mult
    weights[:int(neighborhood_radius),:] *= img_edge_mult
    weights[-int(neighborhood_radius):,:] *= img_edge_mult
    weights[:,:int(neighborhood_radius)] *= img_edge_mult
    weights[:,-int(neighborhood_radius):] *= img_edge_mult
                          
    return weights, eroded_mask


# takes a binary (eroded) mask, identifies connected regions, then dilates them according to given radius. Outputs ID'd dilated mask. Note that the order of dilation matters in some cases.
def dilate_and_ID_mask(eroded_mask,neighborhood_radius=def_neighborhood_radius):
    
    # determine connected regions in eroded mask
    connectedmask = np.ubyte(np.array(eroded_mask,dtype=float))
    (numregions, finlabels, _, _) = cv2.connectedComponentsWithStats(connectedmask)
  
    # create neighborhood mask and integer length scale
    C = circ_region(neighborhood_radius)
    R = int(np.floor(neighborhood_radius))

    IDmask_padded = np.pad(np.zeros(np.shape(finlabels)),((R, R), (R, R)),'constant', constant_values=-1 )
    # pad eventual IDmask to avoid edge issues (for now)

    # cycle through entire image, painting neighborhoods
    for r in range(np.shape(finlabels)[0]):
        for c in range(np.shape(finlabels)[1]):
            if finlabels[r,c]>0: # if this is in a connected region, dilate!
                current_neighborhood = IDmask_padded[r:r+2*R+1,c:c+2*R+1]
                current_neighborhood[C]= finlabels[r,c]

                IDmask_padded[r:r+2*R+1,c:c+2*R+1] = current_neighborhood # label all elements in neighborhood with value
            
    
    return IDmask_padded[R:-R,R:-R]
    


def fractionalize_AOI(masks, AOIs, fraction):

    train_innie,train_total = 0,0
    
    for k,AOI_dummy in enumerate(AOIs):
        train_innie += np.sum((masks[k]>0)*AOI_dummy)
        train_total += np.sum(AOI_dummy)
        
    train_innie_frac = train_innie/train_total
    train_delta = [(fraction/2)/(1-train_innie_frac) - 1  ,  (fraction/2)/train_innie_frac - 1] # [0] is outie, [1] is innie

    print('[BB] Balance classes by over/under sampling: innies x',np.round(train_delta[1]+1,2),' outies x',np.round(train_delta[0]+1,2))
    
    for k in range(len(AOIs)):

        train_delta2 = copy.copy(train_delta)
        checkmask = masks[k] > 0
        AOImask = 0*AOIs[k] + 1

        for inout in [0,1]:

            while train_delta2[inout] >=1: # if small enough, one class may need to be added multiple times
                train_delta2[inout] -=1
                AOImask[checkmask==inout]+=1

            # adjust by remaining delta, so that innies are now fraction/2 of AOI (may be counted 2 or more times per px)
            dummy = np.random.choice([0, 1], size=AOImask.shape, p=(1-abs(train_delta2[inout]), abs(train_delta2[inout])))
            # add to AOI count if positive adjustment, subtract from it if negative (np.sign)
            
            AOImask[checkmask==inout] += (np.sign(train_delta2[inout])*dummy[checkmask==inout]).astype(int)
                                    
        AOIs[k]  = AOImask*AOIs[k]

                                    
    
    return AOIs


  


import configparser
import cv2
import numpy as np
import os
import copy
from skimage.segmentation import watershed


# Credit: https://stackoverflow.com/questions/2068372/fastest-way-to-list-all-primes-below-n-in-python/3035188#3035188
def primesto(n):
    """ Returns a array of primes, 3 <= p < n """
    sieve = np.ones(n//2, dtype=bool)
    for i in range(3,int(n**0.5)+1,2):
        if sieve[i//2]:
            sieve[i*i//2::i] = False
    return np.insert(2*np.nonzero(sieve)[0][1::]+1,0,2)

primelist = primesto(10**7) # this is 664579 primes. If you have more particles than this, good luck.
primelist = np.insert(primelist,0,0)


# create a circular region where pixels within distance r of the center are 1's, others are 0's.
def circ_region(r):
    
    if r <=1:
        return 1
    
    R = int(np.floor(r))
    xv,yv = np.meshgrid(range(-R,R+1),range(-R,R+1))
    
    return (xv**2 + yv**2) <= r**2


def save_parameters(filename,keys,vals):

    config = configparser.RawConfigParser()

    config.add_section('Parameters')
    
    for k,v in zip(keys,vals):
        config.set('Parameters',  k.lower(), v)
        
    with open(filename+'.txt', 'w') as configfile:
          config.write(configfile)
            
def load_parameters(filename,param_types):
    config = configparser.ConfigParser()
    config.read(filename+'.txt')
    
    keys = list(param_types.keys())
    values = list(param_types.values())
    
    out = [0 for _ in keys]
    keys2 = [k.lower() for k in keys]

    for key in config['Parameters']: 
        if key.lower() in keys2:
            out[keys2.index(key.lower())]=config['Parameters'][key.lower()]
    
    param = LoadConverter(out,values)
    param = dict(zip(keys,param))

    return param
    
    
    
#Convert array of parameters formatted as strings into array of parameters formatted according to param_var_types
def LoadConverter(param_string,param_var_types):
    for ind in range(len(param_string)):
        param_string[ind] = param_var_types[ind](param_string[ind])
    return param_string

    
def load_image(filepath, binarize = False, integerize = False, RGB_to_gray = False, segment = False):
    
    if filepath is None:
        return None
    
    if binarize and segment:
        raise Exception('Binarization and Segmentation are incompatible')
    
    
    if filepath[-4:] == '.txt': # load text
        img = np.array(np.loadtxt(filepath,skiprows=0))
    else: # load image file
        img = cv2.imread(filepath, cv2.IMREAD_ANYDEPTH)
        if img is None:
            print('Default image loading method failed, trying backup')
            img = cv2.imread(filepath)
        if img is None:
            raise Exception('Loaded image is empty')
        if len(np.unique(img)) == 1:
            print('Only one value for the entire image '+filepath)
    
    
    # average color channels
    if RGB_to_gray:
        #print('Converting '+ filepath + ' to grayscale before use.')
        if len(np.shape(img))==3 and np.shape(img)[2]==3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        elif len(np.shape(img))>=3:
            print('Error for: '+filepath)
            raise Exception('Attempted to convert to grayscale but image is wrong dimensions')
        
    # convert all non-zero values to 1's
    if binarize:
        if len(np.shape(img))>2 and np.shape(img)[2]>1:
            for k in range(1,np.shape(img)[2]):
                if not np.array_equal(img[:,:,0], img[:,:,k]):
                    raise Exception('Attempting to binarize image with multiple and different color channels.')
            img = img[:,:,0] # reduce to one channel
            
        img = 1*(img!=0)
    
    
    # take given IDs and break them up according to disconnected regions (works for binary masks, too).
    if segment:
        
        # distance inside ID'd regions
        dist = distance_maker(img)

        # give each disconnected region a unique ID
        (_, markers, _, _) = cv2.connectedComponentsWithStats((dist>2).astype('uint8'))

        # watershed using these IDs, fill out given mask, use distance as rate
        img = watershed(-dist, markers = markers, mask=(img>0)).astype(int)
        
        
    
    
    
    
    # convert unique values into unique integers. 0 is interpreted as background (outies)
    if integerize and not segment: # segmenting already does this
        S = np.shape(img)
        if len(S)>2 and S[2]>1:
            for k in range(1,S[2]):
                if not np.array_equal(img[:,:,0], img[:,:,k]):
                    raise Exception('Attempting to integerize image with multiple and different color channels.')
            img = np.sum(img,tuple(k for k in range(2,len(S)))) # reduce to one channel
            
         # take unique values and convert them to indicees
        U = np.unique(img)
        img2 = np.copy(img)
        for idx,val in enumerate(U):
            img[img2==val] = idx+1
            
        img[img2==0] = 0 # maintain background as background (it was assigned a different value above)
        img = np.array(img,dtype=int)
    return img


# find masks from filenames mask_names with matching names (ignore file extension) to the image filenames in img_names
# return a list of mask names that match, in the same order as img_names 
def find_matching_masks(img_names, mask_names, raise_issues=True):
     
    img_name_bases = [name[:name.index('.')] for name in img_names]
    mask_name_bases = [name[:name.index('.')] for name in mask_names]
    mask_idxs = [mask_name_bases.index(img_name) if img_name in mask_name_bases else None for img_name in img_name_bases]
    
    if raise_issues and None in mask_idxs:
        raise Exception('Masks are missing for the following images: '+', '.join([name for name, idx in zip(img_names,mask_idxs) if idx is None]))
    
    return [mask_names[k] if not k is None else None for k in mask_idxs] 
    

# returns a string listing input image names and their sizes
def list_imgs_and_sizes(img_names,imgs):
    # str call makes None type into 'None'
    out = [str(name) + ' ('+','.join(str(k) for k in np.shape(img))+')' for name,img in zip(img_names,imgs)]
    return ', '.join(out)  
    
# returns a list of filenames (assumption is they are all images or text)
# want is indicees of desired images (None -> all files), count is total number of names returned (always a subset of want/all)
def get_image_list(folder, want = None, count=-1):
    
    names = os.listdir(folder) 
    for name in names: # remove .DStore files and the like
        if name[0] == '.':
            names.remove(name)

    # want = set of images to draw from (None means all). count = number of images to end up with (-1 means all)
    if count ==-1: # take all images of want
        if want is None:
            want = np.linspace(0, len(names)-1,len(names))
    
    elif count >= 0: # take [count] images
        if want is None:
            want = np.linspace(0, len(names)-1, count)
        else:
            want = want[np.linspace(0, len(want)-1, count)]

    else: # count needs to be positive or -1
        raise Exception('Image counting parameter has an illegal value.')
  
    # narrow names down only to desired ones, then create array for images
    return [names[int(k)] for k in want] 



# returns images of a list of filenames for a given folder
def load_image_list(folder, names, binarize = False, integerize = False, RGB_to_gray = False, chatter = True, segment = False):
    
    if RGB_to_gray and (not integerize) and (not binarize) and chatter:
        index = folder.rfind('/') # include only this folder's name (not filepath)
        print('[BB] Converting images in '+ folder[index+1:] + ' to grayscale before use.')
        
    names2 = copy.copy(names)
    for i,name in enumerate(names2):
        if not name is None:
            names2[i] = folder + '/' + name
            
    return [load_image(name, binarize = binarize, \
                          integerize = integerize, RGB_to_gray = RGB_to_gray, segment = segment) \
                          for name in names2] 


def load_images_aois_masks(param,img_folder,AOI_folder,mask_folder=None):
    #Find image list
    img_names, imgs = get_image_list(img_folder, RGB_to_gray = param['images_to_grayscale'])
    
    #If there are AOI's, load them
    #if aoi_folder!=None:
    AOI_names = os.listdir(AOI_folder)
    AOI_names = find_matching_masks(img_names, AOI_names, raise_issues=False)
    _, AOIs = get_image_list(AOI_folder, AOI_names, integerize = True)
    AOIs = [elem if (not elem is None) else np.ones(np.shape(imgs[k])[:2]) for k,elem in enumerate(AOIs)]
    
    #If there are masks, load them
    if mask_folder!=None:
        mask_names = os.listdir(mask_folder)
        mask_names = find_matching_masks(img_names, mask_names)
        _, masks = get_image_list(mask_folder,mask_names, integerize = True)
    else:
        mask_names, masks = [],[]
    
    return img_names,imgs,AOI_names,AOIs,mask_names,masks




# creates a 2D distance map (dist to nearest border) from an ID'd mask.
def distance_maker(mask):

    eroder = 1.01
    img2 = np.sqrt(np.array(primelist[np.array(mask,dtype=int)],dtype=float)) # make segmented image using only primes, then take sqrt
    kernel = np.array(circ_region(eroder),dtype=np.uint8)

    # convolve image over kernel -- those that catch another sqrt(prime) will become non-integers when multiplied by sqrt(prime) img
    eroded_mask = np.multiply(cv2.filter2D(img2,-1,kernel),img2) 

    # detect non integers - these are innies next to other particles
    twoParticles = 1-(np.abs(eroded_mask-np.round(eroded_mask))>.000001)  #this number is somewhat arbitrary. Too small is a problem though.
    

    #eroded_mask = cv2.erode(np.array(img2>0,dtype=float),kernel) - twoParticles # erode and subtract particle-particle borders
    #eroded_mask[eroded_mask<0] = 0 # any subtractions below 0 from twoParticles are ignored.
    # mask now is binary and eroded by 1 pixel from ID'd mask

    # take distance transform (distance from zeros)
    dist = cv2.distanceTransform(np.array(mask>0,dtype='uint8'), distanceType=cv2.DIST_L2, maskSize=0)
    
    # take distance from two particle borders (and add one, since border itself is 2px)
    dist2P = 1+cv2.distanceTransform(np.array(twoParticles,dtype='uint8'), distanceType=cv2.DIST_L2, maskSize=0)
    #dist[dist>0]+=1
    #dist += 1*((mask>0) * (eroded_mask==0)) # add back in eroded away spots (with distance 1)
    
    # distance should be capped and also minimum of distance to 0 and distance to another particle
    dist = np.minimum(dist,dist2P)
    
    
    #dist -= cv2.distanceTransform(np.array(1-(mask>0),dtype='uint8'), distanceType=cv2.DIST_L2, maskSize=0)
   # dist[dist<-self.dist_max] = -self.dist_max
                    
    return dist


def create_default_params():

    param = {};

    param['S_half'] = 12# int, defines size of input image to NN: size = (2[S_half]+1 square). 12 is reccommended.

    param['scales'] = 4 # number of scaled images (int, 1 is just regular without scaling)
    param['scalefactor'] = 2 # scale multiplier (int)
    # for example, with 3 scales of scalefactor 4, images scaled by 1, 4^1=4, and 4^2=16 are included as inputs

    param['batch_size'] = 1024 # int, must be multiple of 8 to allow rotations + flips. Same size for testing.
    #True/False
    param['rotations'] = 1 # rotate images by 90, 180, 270 degrees in training to multiply data by 4
    param['flips'] = 1 # flip images in training to multiply data by 2

    #True/False
    param['images_to_grayscale'] = 1; #convert images to grayscale (if images already one channel, this has no effect.)

    param['HP_network_num'] = 7; # For a list of what networks these numbers create, see readme.
    param['HP_train_epochs'] = 3; # how many epochs to train

    # (total AOI area)*fraction = training samples 
    # These are automatically adjusted such that half come from each class, meaning some may be multi-counted.
    # This does not affect test set.
    param['fraction'] = 0.3

    # maximum distance to train into network (dist from nearest outie). 2*dist_max < min particle diameter
    param['dist_max'] = 10;  

    param['track_outies']  = 0 # invert inputs and track 'outies' instead (works for small, separated particles)

    # optional values for spatially weighting pixels near borders of particles/images
    param['HP_neighborhood_radius'] = 1 # neighborhood of relevance for weighting
    param['HP_particle_border_mult'] =  1# weight multiplier for pixels within HP_neighborhood of a particle border
    param['HP_two_particle_mult'] =1 # weight multiplier for pixels within HP_neighborhood of two particles
    param['HP_img_edge_mult'] = 1 # weight multiplier for pixels within HP_neighborhood of an image edge

    # Specify the type(s) of final images to be saved, True(1) or False(0)
    param['output_segmented']=1
    param['output_binarized']=0
    param['output_dist']=0
    param['output_classprob']=0
    param['output_markers']=0

    # Pick output file type (.png / .npy)
    param['save_to_png']=1
    param['save_to_npy']=0
    
    return param


from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Conv2D, Flatten, Dropout, MaxPooling2D
from tensorflow.keras.backend import clear_session
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import layers, Input, Model

def model_S_half_minimum(model_num):
    
    if model_num == 7:
        return 9
    else:
        raise Exception('Model Number Not Recognized.')
        

# Create a keras sequential NN that will train using the input train_gen (training data generator). 
def generate_network(input_shape,model_num=7):
    
    clear_session()

    #create model
    model = Sequential()
    
    
    #add model layers
    if not input_shape[0]>= 2*model_S_half_minimum(model_num):
        raise Exception('Model '+str(model_num)+' requires S_half >= ',str(model_S_half_minimum(model_num)),'.')
    
    if model_num == 7:

        inputs = Input(shape=input_shape, name="img")
        x = layers.Conv2D(64, 3, activation="relu")(inputs)
        x = layers.Conv2D(32, 3, activation="relu")(x)           
        block_1_output = layers.MaxPooling2D(pool_size=(2,2))(x)

        x = layers.Conv2D(32, 3, activation="relu", padding="same")(block_1_output)
        x = layers.Dropout(.1)(x)
        x = layers.Conv2D(32, 3, activation="relu", padding="same")(x)
        block_2_output = layers.add([x, block_1_output])
        block_2_output = layers.MaxPooling2D(pool_size=(2,2))(block_2_output)
                   
        x = layers.Conv2D(32, 3, activation="relu", padding="same")(block_2_output)
        x = layers.Dropout(.1)(x)
        x = layers.Conv2D(32, 3, activation="relu", padding="same")(x)
        block_3_output = layers.add([x, block_2_output])
        block_3_output = layers.MaxPooling2D(pool_size=(2,2))(block_3_output)

        x = layers.Flatten()(block_3_output)
        x = layers.Dense(64, activation="relu")(x)
        x = layers.Dropout(.1)(x)
        x = layers.Dense(64, activation="relu")(x)
        x = layers.Dropout(.1)(x)
        x = layers.Dense(64, activation="relu")(x)
        x = layers.Dropout(.1)(x)
        x = layers.Dense(64, activation="relu")(x)
        x = layers.Dropout(.1)(x)
        outputs2 = layers.Dense(1,activation="relu", name="distance_output")(x)
        
        x2 = layers.add([x,outputs2])
        outputs = layers.Dense(2,activation="softmax",name="category_output")(x2)


        model = Model(inputs, [outputs,outputs2], name="BB_resnet_2out")
        
       # two losses
        losses = {
        "category_output": "categorical_crossentropy",
        "distance_output": "mean_absolute_error",
        }
        lossWeights = {"category_output": 1.0, "distance_output": 1.0}

        opt = Adam(learning_rate = 0.001)
        
        model.compile(optimizer=opt, loss=losses, loss_weights=lossWeights, metrics=['accuracy'])

        return model
    
    else:
        raise Exception('Model number not recognized')
        
        
    
    


#######################################################################################
import tensorflow as tf
import numpy as np
import os,sys,datetime
import gc
from pathlib import Path
from DataGenerator import circ_region, CreateGenerator, HPinputGenerator
from saved_model_functions import convert_param_to_outputs,  CreateModel, Predict
from generate_network import model_S_half_minimum
from load_save_helpers import load_parameters,save_parameters

#sys.path.append(os.path.dirname(os.path.realpath(__file__))) # Add parent directory to sys.path (for script)
sys.path.append(os.path.dirname(os.getcwd())) # Add parent directory to sys.path (for notebook)

# round results (e.g. accuracy) to this many decimal places
round_out = 5

# remove test set from fit function. Speeds up training, loses mid-training updates 
include_test_in_training = False



physical_devices = tf.config.list_physical_devices('GPU') 

if len(physical_devices):
    #print('LIMITING TO 1 GPU')
    tf.config.set_visible_devices(physical_devices[-1:], 'GPU')

if len(physical_devices):
   # print('SETTING MEMORY GROWTH ON GPU TO TRUE')
    tf.config.experimental.set_memory_growth(physical_devices[0], True)

####### END OF HEADER #################################################################


param_types={'S_half':int, 'scales':int, 'scalefactor':int, 'HP_neighborhood_radius':float, 'HP_particle_border_mult':float, 'HP_two_particle_mult':float, 'HP_img_edge_mult':float, 'batch_size':int, 'rotations':int, 'flips':int, 'images_to_grayscale':int, 'HP_network_num':int, 'HP_train_epochs':int, 'fraction':float, 'output_segmented':int, 'output_binarized':int, 'output_dist':int, 'output_classprob':int, 'output_markers':int, 'save_to_png':int, 'save_to_npy':int,'track_outies':int,'dim3':int, 'dist_max':int}

norms_types = {'img_min':float,'img_max':float}

results = ['test_acc','train_acc','test_right','test_total','train_right','train_total','train_innie_f1_preseg','train_outie_f1_preseg','test_innie_f1_preseg','test_outie_f1_preseg','train_innie_f1','train_outie_f1','test_innie_f1','test_outie_f1','train_seg','test_seg','train_area_weighted_seg','test_area_weighted_seg']
res_types = dict(zip(results,[float for _ in results]))




def BBHP(filepath,param,train_img_count=-1,test_img_count=-1,want_train=None, want_test=None,dt_string=None,predict_path=None):
    
    
    #######################################################################################
    # INDICATE ALL FOLDERS WHERE IMAGES, MASKS, AND AOI ARE STORED
    #######################################################################################
    
    # subfolders:
    train_img_folder = filepath+'train_images'      # images for training (with masks in mask_folder)
    test_img_folder = filepath+'test_images'        # images for testing (with masks in mask_folder)
    predict_img_folder = filepath+'predict_images'  # images for prediction (no masks needed)
    mask_folder = filepath+'masks'                  # ID'd masks with 0 as outies. Must be same 2D shape as training and test images
    AOI_folder = filepath+'areas_of_interest'       # binary masks representing area of interest (only 1s are included in training, testing, and predicting.)
    # if AOI does not exist for a given image, it is assumed every pixel is in the area of interest.

    # NOTE: masks, images, and AOIs are matched by name. The extension can be different, 
    # but the filename prior to the extension should be identical.

    # The max and min pixel values of the entire training set will be measured, and then all images
    # will be scaled so that these values correspond to 0.5 and -0.5 respectively. 
    # These scalings are not saved in the image files, but the max/min is stored in the "normalization.txt" file.


    #### CHECK PARAMETER COMBINATIONS, SAVE TO TEXT FILE AND LOAD THEM INTO WORKSPACE VARIABLES
    #######################################################################################
    
    norms = None
    param_input = param.copy()
    if dt_string is not None:
        outputfilename = filepath+'outputs_'+dt_string
        print('Date string passed, retrieving parameters and normalizations from ' + outputfilename)
        parameter_filename = outputfilename+'/parameters'
        HP_weight_filename = outputfilename+'/networkweights.ckpt'
        normalization_filename = outputfilename+'/normalization'
        
        
        param = load_parameters(parameter_filename,param_types) # load old parameters
        old_epochs = param['HP_train_epochs']
        
        #Adjust loaded (old) parameters to match any input (new) parameters
        for key,value in param_input.items():
            param[key] = value
            
        num_epochs = param_input['HP_train_epochs'] # must have specified this value in new input

        norms = load_parameters(normalization_filename,norms_types)
        
        param['HP_train_epochs'] += old_epochs # store the TOTAL number of epochs trained
    
        if num_epochs>0:
            
            dt_string = datetime.datetime.now().strftime("%y_%m_%d_%H_%M_%S")
            outputfilename = outputfilename+'_'+dt_string
            index = outputfilename.rfind('/')
            
            print('[BB] More training requested, creating '+outputfilename[index+1:]+' to hold new outputs')
            os.mkdir(Path(outputfilename))
            
            parameter_filename = outputfilename+'/parameters'
            normalization_filename = outputfilename+'/normalization'
            #HP_weight_filename = outputfilename+'/networkweights.ckpt'
        

        else:
            print("[BB] No more training requested, outputs staying in original folder")
    
    

    else:
        # create current date/time string that will be used to catalog all outputs and parameters
        dt_string = datetime.datetime.now().strftime("%y_%m_%d_%H_%M_%S")
        outputfilename = filepath+'outputs_'+dt_string
        print('[BB] Outputs will be stored in outputs_'+dt_string)
        os.mkdir(Path(outputfilename))

        
        parameter_filename = outputfilename+'/parameters'
        HP_weight_filename = None#outputfilename+'/networkweights.ckpt'
        normalization_filename = outputfilename+'/normalization'


        # Training batch size must be correct multiple when using rotations/flips
        unique_in_batch = param['batch_size']/((1+3*param['rotations'])*(1+param['flips']))
        if not unique_in_batch == int(unique_in_batch):
            raise Exception('Batch number must be divisible by 4 if doing rotations, 2 if flipping, and 8 if both.')

        # input window size must be big enough for selected network
        if not param['S_half'] >= model_S_half_minimum(param['HP_network_num']):
            raise Exception('HP network model '+str(param['HP_network_num'])+ ' requires S_half >= '+str(model_S_half_minimum(param['HP_network_num']))+'.')
        
        num_epochs = param['HP_train_epochs']
        
        
    if num_epochs>0:
        #print('[DEBUG] Preparing generators for model training...')
        
        #### LOAD TRAIN/TEST IMAGES, MASKS, AOIs, CREATE DATAGENERATORS FOR TRAINING/TESTING
        #######################################################################################

        train_genHP,train_img_names,_ = CreateGenerator(filepath=filepath, subfolder='train_images',\
                                      param=param, want=want_train, \
                                      count=train_img_count, train_test_predict=0, chatter = True)

        if include_test_in_training:
            test_genHP,test_img_names,_ = CreateGenerator(filepath=filepath, subfolder='test_images',\
                                         param=param, want=want_test, \
                                         count=test_img_count, train_test_predict=1, chatter = True)
        else:
            test_genHP = None
    
        #Add number of channels in image to parameters
        param['dim3'] = train_genHP.dim3
        if norms is None:
            
            #### Generate Normalization values from training images
            #######################################################################################
            norms = {}
            norms['img_max'],norms['img_min'] = train_genHP.get_img_range()
        
        #Normalize the images in our generators   
        train_genHP.normalize_images(norms['img_max'],norms['img_min'])
        if include_test_in_training:
            test_genHP.normalize_images(norms['img_max'],norms['img_min'])
            
    elif num_epochs==0:
        print('Number of epochs = 0, entering predict only mode')
        train_genHP = None
        test_genHP = None
    elif num_epochs<0:
        raise Exception('Please change number of epochs to be >=0')
    
    # save parameters
    save_parameters(parameter_filename,list(param.keys()),list(param.values()))
    save_parameters(normalization_filename,list(norms.keys()),list(norms.values()))

    #### CREATE AND TRAIN NEURAL NETWORK FOR HP, SAVE FINAL WEIGHTS, GENERATE PREDICTIONS
    #######################################################################################

    
    #train_genHP.plot_items(5)
    
    print('[BB] Training Neural Network')
    if not include_test_in_training:
        print('[BB] -- Excluding test images from .fit function. Test set results available at end.')

    modelHP = CreateModel(filepath,dt_string,num_epochs,param,outputfilename,train_gen=train_genHP,test_gen=test_genHP,chkpt_filepath=HP_weight_filename)
    
    # garbage collect
    del train_genHP
    del test_genHP
    gc.collect()
    
    #If dt_string has changed during training, update receiving folders for output
    HP_output_folder = outputfilename+'/predictions'
    results_filename = outputfilename+'/results'
        
    if not (param['save_to_png'] or param['save_to_npy']):
        print('No save flags are activated, prediction images/data will not be saved')
        
    if num_epochs>0:
        os.mkdir(HP_output_folder)
        results_filename = outputfilename+'/results'

        # generator for training data
        #print('[BB] Calculating accuracy and exporting images of predictions...')
        train_true_pred = np.zeros((2,2));
        train_true_pred_seg = np.zeros((2,2));
        train_SEG = [];
        train_areas = [];
        test_SEG = [];
        test_areas = [];
        more_data = True
        index = 0
        
        images_to_save = convert_param_to_outputs(param)
        print('[BB] Exporting',', '.join([i[1:] for i in images_to_save]))
        
        while more_data:
            try:
                all_genHP,img_names,seg_masks = CreateGenerator(filepath=filepath, subfolder='train_images',\
                                      param=param, want=want_train, \
                                      count=train_img_count, train_test_predict=1,index = index, chatter = False)
                all_genHP.normalize_images(norms['img_max'],norms['img_min'])
                SEG0, areas0 = Predict(param, modelHP, all_genHP, HP_output_folder, img_names, true_pred = train_true_pred, true_pred_seg = train_true_pred_seg,seg_masks = seg_masks)
                
                train_SEG += SEG0
                train_areas += areas0
                
                print('[BB] -- Train Image '+str(index)+': '+img_names[0]+' -- SEG: '+str(np.round(np.mean(SEG0),round_out)))

                    
                index+=1
                del all_genHP
                gc.collect()
            except IndexError:
                more_data= False

        # generator for test data
        test_true_pred = np.zeros((2,2))
        test_true_pred_seg = np.zeros((2,2))
        more_data = True
        index = 0
        while more_data:
            try:
                all_genHP,img_names,seg_masks = CreateGenerator(filepath=filepath, subfolder='test_images',\
                                      param=param, want=want_train, \
                                      count=train_img_count, train_test_predict=1,index = index)
                all_genHP.normalize_images(norms['img_max'],norms['img_min'])
                SEG0, areas0 = Predict(param, modelHP, all_genHP, HP_output_folder, img_names, true_pred = test_true_pred, true_pred_seg = test_true_pred_seg, seg_masks = seg_masks)
                
                test_SEG += SEG0
                test_areas += areas0
               
                print('[BB] -- Test Image '+str(index)+': '+img_names[0]+' -- SEG: '+str(np.round(np.mean(SEG0),round_out)))

                index+=1

                del all_genHP
                gc.collect()
            except IndexError:
                more_data= False

        tf.keras.backend.clear_session()
        train_f1s = f1_score(train_true_pred)
        test_f1s = f1_score(test_true_pred)    

        train_f1s_seg = f1_score(train_true_pred_seg)
        test_f1s_seg = f1_score(test_true_pred_seg)

        #print(train_true_pred)
        #print(test_true_pred)

        results = {}
        results['test_acc'] = np.sum(np.diag(test_true_pred_seg))/np.sum(test_true_pred_seg)
        results['train_acc'] = np.sum(np.diag(train_true_pred_seg))/np.sum(train_true_pred_seg)
        results['test_right'] = np.sum(np.diag(test_true_pred_seg))
        results['test_total'] = np.sum(test_true_pred_seg)
        results['train_right'] = np.sum(np.diag(train_true_pred_seg))
        results['train_total'] = np.sum(train_true_pred_seg)



        results['train_innie_f1_preseg'] = train_f1s[1]
        results['train_outie_f1_preseg'] = train_f1s[0]
        results['test_innie_f1_preseg'] = test_f1s[1]
        results['test_outie_f1_preseg'] = test_f1s[0] 
        results['train_innie_f1'] = train_f1s_seg[1]
        results['train_outie_f1'] = train_f1s_seg[0]
        results['test_innie_f1'] = test_f1s_seg[1]
        results['test_outie_f1'] = test_f1s_seg[0] 
        
        

        results['train_SEG'] = np.mean(train_SEG)
        results['test_SEG'] = np.mean(test_SEG)
        results['train_area_weighted_SEG'] = np.mean(np.array(train_SEG)*np.array(train_areas))/np.mean(train_areas)
        results['test_area_weighted_SEG'] = np.mean(np.array(test_SEG)*np.array(test_areas))/np.mean(test_areas)

       
        for key in results:
            results[key] = round(results[key],round_out )
        
        print(results)


        # save results
        save_parameters(results_filename,list(results.keys()),list(results.values()))

    elif num_epochs==0:
        if os.path.isdir(HP_output_folder)==False:
            print('Original training never made predictions, creating prediction folder now...')
            os.mkdir(HP_output_folder)

        # generator for prediction data
        if predict_path is None or predict_path == 'predict_images':
            subfolder = 'predict_images'
            print('Predicting on images in predictions folder...')
            true_pred = None
            true_pred_seg = None
            train_test_predict = 2
            prestr = '';
        elif predict_path == 'test_images':
            subfolder = 'test_images'
            print('Predicting on images in the test folder...')
            true_pred = np.zeros((2,2))
            true_pred_seg = np.zeros((2,2))
            train_test_predict = 1
            prestr = 'test_'
        elif predict_path == 'train_images':
            subfolder = 'train_images'
            print('Predicting on images in the train folder...')
            true_pred = np.zeros((2,2))
            true_pred_seg = np.zeros((2,2))
            train_test_predict = 1
            prestr = 'train_'
        else:
            raise Exception("predict_path must be None (defaults to predict_images), 'train_images', 'test_images', or 'predict_images'")
        more_data = True
        index = 0
        prediction_results_filename = outputfilename+'/prediction_results'
        while more_data:
            try:
                all_genHP,img_names,seg_masks = CreateGenerator(filepath=filepath, subfolder=subfolder,\
                                      param=param, train_test_predict=train_test_predict,index = index, chatter = False)
                all_genHP.normalize_images(norms['img_max'],norms['img_min'])

                SEG0, areas0 = Predict(param, modelHP, all_genHP, HP_output_folder, img_names,true_pred = true_pred, true_pred_seg = true_pred_seg ,seg_masks = seg_masks)
                index+=1
                del all_genHP
                gc.collect()
            except IndexError:
                more_data= False
        if true_pred is not None:
            tf.keras.backend.clear_session()
            f1s = f1_score(true_pred)
            f1s_seg = f1_score(true_pred_seg)

            results = {}
            results[prestr+'acc'] = np.sum(np.diag(true_pred_seg))/np.sum(true_pred_seg)
            results[prestr+'right'] = np.sum(np.diag(true_pred_seg))
            results[prestr+'total'] = np.sum(true_pred_seg)



            results[prestr+'innie_f1_preseg'] = f1s[1]
            results[prestr+'outie_f1_preseg'] = f1s[0]
            results[prestr+'innie_f1'] = f1s_seg[1]
            results[prestr+'outie_f1'] = f1s_seg[0]

            for key in results:
                results[key] = round(results[key],round_out )
        
            print(results)


            # save results
            save_parameters(prediction_results_filename,list(results.keys()),list(results.values()))

    return dt_string

# takes square prediction matrix and outputs f1 score for each class
def f1_score(true_pred):
    out = np.zeros(len(true_pred))
    for k in range(len(true_pred)):
        # precision = true_positives / predicted_positives
        precision = true_pred[k,k]/np.sum(true_pred[:,k])
        # recall = true_positives / possible_positives
        recall = true_pred[k,k]/np.sum(true_pred[k,:])
        out[k] = 2*((precision*recall)/(precision+recall+0.0000000001))
    
    return out


import tensorflow as tf
import numpy as np
import cv2
from skimage.segmentation import watershed
from DataGenerator import circ_region
from generate_network import generate_network
from scipy.ndimage import maximum_filter

#Dictionary key for text and variable translation for image_dict, used in SegmentPrediction, UpdateAccTable, SaveImage and SaveMatrix functions
possible_prediction_names = {'output_binarized':'outmask','output_classprob':'mask','output_dist':'dist','output_segmented':'labels','output_markers':'markers'}

#Dictionary for scaling prediction outputs for image values, used in SaveImage
pixel_scaling = {'_binarized': 255,'_classprob': 255,'_dist': 50,'_segmented': 97,'_markers': 97}

def convert_param_to_outputs(param):
    possible_prediction_names = {'output_binarized':'outmask','output_classprob':'mask','output_dist':'dist','output_segmented':'labels','output_markers':'markers'}
    to_save = []
    #Iterate through param variables
    for key in param:
        #Find the variables related to output formatting
        if key in possible_prediction_names.keys():
            #Select only the desired formats for output
            if param[key]:
                #remove the 'output' tag from the string and add to final list
                to_save.append(key[6:])
    return to_save


#Given a prediction and error functions, segment and watershed the prediction while updating the error count
def SegmentPrediction(BBout,dist_max):
    '''
    inputs:
    BBout = single prediction from BBpredictions
    dist_max = parameter from the generator, max value network predicts for distance to outie
    outputs:
    image_dict = dictionary containing segmented matrix and any other desired outputs
    ASSUME BACKGROUND IS 0
    '''
    kernel = np.array(circ_region(dist_max),dtype=np.uint8)
    dist = BBout[:,:,1]
    mask = BBout[:,:,0]
    outmask = BBout[:,:,0]>0.5

    # find points that are highest in their neighborhood
    maxima = dist == maximum_filter(dist, footprint=kernel) 
    maxima = np.array(maxima*outmask, dtype=np.uint8)
    maxima[dist<dist_max/2] = 0
    
    # Old way of doing things. Obsolete.
    if False:
        maxima[dist>dist_max-1] = 1

        #connectedmask = np.ubyte(np.array(cv2.filter2D(dist,-1,kernel)>=(np.sum(kernel)*(dist_max-0.5)),dtype=float))
        (_, markers, _, _) = cv2.connectedComponentsWithStats(maxima)

        labels = watershed(-dist, markers = markers, mask=outmask)

    if True:

        # give each a unique ID
        #connectedmask = np.ubyte(np.array(cv2.filter2D(dist,-1,kernel)>=(np.sum(kernel)*(dist_max-0.5)),dtype=float))
        (_, markers, _, _) = cv2.connectedComponentsWithStats(maxima)

        # watershed using these maxima, only on classprob>0.5, and using distance as rate
        labels = watershed(-dist, markers = markers, mask=outmask)
        
        # watershedding creates too many segments, find those that should be combined
        S = np.shape(labels)
        combinemat = np.zeros((np.max(labels)+1,np.max(labels)+1)) # catalog which regions should be combined

        # for each pixel jump right or down, if sum of dist of both pixels is too high, these regions are the same
        for r in range(0,S[0]-2):
            for c in range(0,S[1]-2):
                if labels[r,c]!=0: 
                    if labels[r,c+1] !=0 and labels[r,c+1]!=labels[r,c]:
                        combinemat[labels[r,c],labels[r,c+1]] += (dist[r,c]+dist[r,c+1])>(1.5*(dist_max))
                    if labels[r+1,c] !=0 and labels[r+1,c]!=labels[r,c]:
                        combinemat[labels[r,c],labels[r+1,c]] += (dist[r,c]+dist[r+1,c])>(1.5*(dist_max))

        # don't know which was first
        combinemat += np.transpose(combinemat)

        # relabel lower numbered regions with higher number if combining is needed
        S2= np.shape(combinemat)
        for r1 in range(1,S2[0]):
            for r2 in range(r1+1,S2[1]):
                if combinemat[r1,r2]: # combine these regions
                    labels[labels==r1]=r2
                    combinemat[r2,:] += combinemat[r1,:] # new combined region inherits all connections of merged regions

        # relabel with continuous values from 1            
        _, labels = np.unique(labels, return_inverse=True) 
        labels = np.array(np.reshape(labels,S),dtype='int32')

    
    #Add manipulated matrices to dictionary
    image_dict = {}
    image_dict['_binarized'] = outmask
    image_dict['_classprob'] = mask
    image_dict['_dist'] = dist
    image_dict['_segmented'] = labels
    image_dict['_markers'] = markers
    
    return image_dict
    

#Given ground truth and segmented predictions, update accuracy tables
#def UpdateAccTable(image_dict,true_pred,true_pred_seg,mask,AOI):
#    '''
#    image_dict = dictionary of final images from SegmentPrediction
#    true_pred = accuracy table from binarized prediction
#    true_pred_seg = accuracy table from segmented prediction
#    mask = Ground truth mask for image
#    AOI = AOI for image
#    '''
#    outmask = image_dict['_binarized']
#    labels = image_dict['_segmented']
#    
#    for inout in [0,1]:
#        for inout2 in [0,1]:
#            true_pred[inout,inout2] += np.sum((outmask==inout2)*((mask>0)==inout)*(AOI>0))
#            
#            true_pred_seg[inout,inout2] += np.sum(((labels>0)==inout2)*((mask>0)==inout)*(AOI>0))

                
#Save segmented matrix as image 
#COME BACK AND CLEAN UP THIS KEY THING ITS STUPID
def SaveImage(BB_output_folder,name,image_dict,images_to_save,dist_max):
    pixel_scaling = {'_binarized': 255,'_classprob': 255,'_dist': 255.0/(dist_max*1.2),'_segmented': 97,'_markers': 97}
    for key in images_to_save:
        cv2.imwrite(BB_output_folder+'/'+name[:name.index('.')]+key+'.png',np.array((pixel_scaling[key]*image_dict[key]) % 256,dtype=int))


#Save segmented matrix as binary .npy           
def SaveMatrix(BB_output_folder,name,image_dict,images_to_save):
    for key in images_to_save:
        np.save(BB_output_folder+'/'+name[:name.index('.')]+key,image_dict[key])
        
        
#Load model parameters from a filepath, institute a model, load weights from a checkpoint (if any) and perform any
#additional training required using the train_gen and test_gen. Save new information to new output path? Add number of 
#epochs already completed as a stored value in a filename?
def CreateModel(filepath,dt_string,num_epochs,param,outputfilename,train_gen=None,test_gen=None,chkpt_filepath = None):    
    
    #Create model
    S = param['S_half']*2+1
    img_shapes = [S,S,param['scales']*param['dim3']]
    modelBB = generate_network(img_shapes,model_num=param['HP_network_num']) 
    
    BB_weight_filename = outputfilename+'/networkweights.ckpt'
    
    
    if chkpt_filepath is None:
        #print("[DEBUG] No checkpoint filename passed, starting from scratch")
        pass
    
    else:
        #This isn't the right file to check for
        #if os.path.isfile(chkpt_filepath)==False:
        #    raise Exception('No checkpoint found at given path')
        print('[BB] Loading model weights from previous checkpoint')
        #Load model
        modelBB.load_weights(chkpt_filepath)
        
        
    #Train the model more if desired
    if num_epochs!=0:
        if train_gen is None:
            raise Exception('[DEBUG] THERE IS NO GENERATOR OF TRAINING DATA TO TRAIN ON')
       # if test_gen is None:
       #     print("[DEBUG] No test generator was passed, so we assume there is none")
            
        # Create a callback that saves the model's weights every epoch
        cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath=BB_weight_filename, verbose=1, save_weights_only=True)
    
        modelBB.fit(train_gen, validation_data=test_gen, epochs=num_epochs,callbacks=[cp_callback]);
    return modelBB
        



#Calculate the SEG accuracy for a whole image vs a ground truth mask.
#Returns np array of Jacard values for each true region.
def segmented_accuracy(mask,prediction,bg_val=0):
    '''
    mask = true segmented values
    prediction = predicted segmented values
    bg_val = value of the region we call background
    '''    

    # shape unimportant now, just overlap.
    mask = mask.flatten()
    prediction = prediction.flatten()

    # find areas of each unique predicted region (store for Jaccard similarity index (J) calculation later)
    pred_regs, areas_pred_regions = np.unique(prediction, return_counts=True)
    
    # get unique values of true regions
    true_regs = list(set(mask))
    true_regs.remove(bg_val) # background not relevant

    SEG = np.zeros(len(true_regs)) # SEG score for each true region
    areas_true_regions = np.zeros(len(true_regs), dtype=int) # Area of each true region


    # Cycle through true regions, find overlaps, find largest overlap and keep Jaccard (if overlap > half)
    for reg_ind,reg in enumerate(true_regs):

        # isolate entire true region in predicted segmentation, store total area, find unique predicted regions within
        pred_regs_local = prediction[mask==reg] # isolate true region
        R = len(pred_regs_local) # calculate its area
        areas_true_regions[reg_ind] = R 
        unique_local_regs, local_areas = np.unique(pred_regs_local, return_counts=True) # list overlapping predicted regions

        max_overlap_idx = np.argmax(local_areas) # find predicted region with highest overlap

        # if this overlap is not background AND is more than half of the true region, calculate J
        if unique_local_regs[max_overlap_idx] != bg_val and local_areas[max_overlap_idx] > R*(0.5):
            I = local_areas[max_overlap_idx] # intersection of pred and real
            P = areas_pred_regions[np.searchsorted(pred_regs,unique_local_regs[max_overlap_idx])] # whole area of pred region
            # R = whole area of real region

            SEG[reg_ind] = I/(P+R-I)

        # if no overlap is high enough, SEG[reg_ind] = 0

    return SEG, areas_true_regions
  
# THIS WILL STAY, BUT IS NOT YET FINISHED.
# Predict function that takes in a generator and model, returns a prediction
def Predict(param, model, gen, BB_output_folder, names, true_pred = None, true_pred_seg = None, seg_masks = None):
    '''
    inputs:
    param: list of parameters 
        including save_to_png and save_to_npy which specify file type to save
        and outputs to save: '_segmented', '_binarized', '_dist', '_classprob', '_markers' (seeds for watershedding)
    model: trained model used for prediction
    gen: generator of images for prediction
    BB_output_folder: filepath for saving segmented predictions 
    names: list of original image names for saving predictions
    true_pred: 2x2 accuracy matrix to be updated on binarized prediction of test set, default None
    true_pred_seg: 2x2 accuracy matrix to be updated on segmented prediction of test set, default None
    seg_masks: ground truth segmented masks
    '''

    BBpredictions = gen.get_prediction(model)
    dist_max = gen.dist_max

    images_to_save = convert_param_to_outputs(param)
    #print('[BB] Exporting ',images_to_save)
    
    SEG = None;
    areas = None; # return none if no mask present
    
    for k, BBout in enumerate(BBpredictions):
        image_dict = SegmentPrediction(BBout,dist_max)
        if (true_pred is not None) and (true_pred_seg is not None):
            mask = gen.get_masks(k)
            AOI = gen.get_AOI(k)
            if len(np.unique(mask))>1:
                #UpdateAccTable(image_dict,true_pred,true_pred_seg,mask,AOI)
                '''
                image_dict = dictionary of final images from SegmentPrediction
                true_pred = accuracy table from binarized prediction
                true_pred_seg = accuracy table from segmented prediction
                mask = Ground truth mask for image
                AOI = AOI for image
                '''
                outmask = image_dict['_binarized']
                labels = image_dict['_segmented']

                for inout in [0,1]:
                    for inout2 in [0,1]:
                        true_pred[inout,inout2] += np.sum((outmask==inout2)*((mask>0)==inout)*(AOI>0))

                        true_pred_seg[inout,inout2] += np.sum(((labels>0)==inout2)*((mask>0)==inout)*(AOI>0))

                SEG,areas = segmented_accuracy(seg_masks[k]*AOI,labels*AOI,bg_val=0) # returns Jacard indicees and areas
               
        if param['save_to_png']:
            SaveImage(BB_output_folder,names[k],image_dict,images_to_save,param['dist_max'])
        if param['save_to_npy']:
            SaveMatrix(BB_output_folder,names[k],image_dict,images_to_save)
        
    return list(SEG),list(areas) # return Jacard indicees for each 


#######################################################################################
import tensorflow as tf
import numpy as np
import os,time,sys,copy,datetime
from BBHP_function import BBHP

sys.path.append(os.path.dirname(os.path.realpath(__file__))) # Add parent directory to sys.path (for script)
#sys.path.append(os.path.dirname(os.getcwd())) # Add parent directory to sys.path (for notebook)

# define useful variables
param = {} # dictionary that holds all parameters
BBfolder = os.path.abspath(os.getcwd()) 
BBfolder = BBfolder[:BBfolder.index('Bellybutton')+len('Bellybutton')+1]+'/' # Bellybutton folder path
####### END OF HEADER #################################################################

### USER INSTRUCTIONS:

#######################################################################################
# (1) INDICATE MAIN FILEPATH WHERE IMAGES ARE
#######################################################################################

# file path where all subfolders below are located
#filepath = BBfolder + 'Useable_Datasets/GFPGOWT1_mouse_stem_cells/'
#filepath = BBfolder + 'Useable_Datasets/3D_Printed_Disks_1/'
filepath = BBfolder + 'UseableDatasets/printed_foam/'


#######################################################################################
# (2) DEFINE ALL RELEVANT PARAMETERS FOR TRAINING
#######################################################################################

param['S_half'] = 12# int, defines size of input image to NN: size = (2[S_half]+1 square). 12 is reccommended, 9 is minimum

param['scales'] = 2 # number of scaled images (int, 1 is just regular without scaling)
param['scalefactor'] = 2 # scale multiplier (int)
# for example, with 3 scales of scalefactor 4, images scaled by 1, 4^1=4, and 4^2=16 are included as inputs

param['batch_size'] = 1024 # int, must be multiple of 8 to allow rotations + flips. Same size for testing.
param['rotations'] = True # rotate images by 90, 180, 270 degrees in training to multiply data by 4
param['flips'] = True # flip images in training to multiply data by 2

param['images_to_grayscale'] = True; #convert images to grayscale (if images already one channel, this has no effect.)

param['HP_train_epochs'] = 3; # how many epochs to train

# (total AOI area)*fraction = training samples 
# These are automatically adjusted such that half come from each class, meaning some may be multi-counted.
# This does not affect test set.
param['fraction'] = 1



param['HP_network_num'] = 7; # For a list of what networks these numbers create, see readme.



#######################################################################################
# Run Bellybutton!
#######################################################################################


BBHP(filepath,param)


